library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity Lane_block is
    Port (  chaining_value : in  STD_LOGIC_VECTOR (255 downto 0);
            message_block : in  STD_LOGIC_VECTOR (511 downto 0);
            rst, clk : in STD_LOGIC;
			start : in  STD_LOGIC;
            done : out  STD_LOGIC;
			bit_counter : in STD_LOGIC_VECTOR(63 downto 0);
			hash_out: out STD_LOGIC_VECTOR(255 downto 0));
end Lane_block;

architecture Behavioral of Lane_block is

--component declarations
component round is
    Port (  plaintext : in  STD_LOGIC_VECTOR (255 downto 0);
            ciphertext : out  STD_LOGIC_VECTOR (255 downto 0);
            rst, clk : in STD_LOGIC;
			start : in  STD_LOGIC;
            done : out  STD_LOGIC;
			bit_counter : in STD_LOGIC_VECTOR(63 downto 0);
			section_number : in STD_LOGIC_VECTOR(2 downto 0);
			rounds : in  STD_LOGIC);
end component;

component Expansion is
	port(	data_in: in std_logic_vector(511 downto 0);
			init_hash: in std_logic_vector(255 downto 0);
			W0, W1, W2, W3, W4, W5: out std_logic_vector(255 downto 0));
end component;

component mux256 is
	port(	a, b, c, d, e, f, g: in STD_LOGIC_VECTOR(255 downto 0);
			sel: in STD_LOGIC_VECTOR(2 downto 0);
			z: out STD_LOGIC_VECTOR(255 downto 0));
end component;

component reg256 is
	port(	reg_in: in STD_LOGIC_VECTOR(255 downto 0);
			clk: in STD_LOGIC;
			en: in STD_LOGIC;
			reg_out: out STD_LOGIC_VECTOR(255 downto 0));
end component;

--signal declarations
signal w: std_logic_vector(1535 downto 0); 				--waarde na boodschapsexpansie 6x256bit
signal done_lane: std_logic;
signal done_round: std_logic;
signal start_round: std_logic;
signal rounds: std_logic;
signal mux_select: std_logic_vector(2 downto 0);
signal mux_out: std_logic_vector(255 downto 0);
signal triexor: std_logic_vector(255 downto 0);			--de ge-exorde waarde van de output van de (P0, P1 en P2) OF (P3, P4 en P5) permutatie
signal permut_out: std_logic_vector(255 downto 0);
signal reg_out: std_logic_vector(1023 downto 0);		--de output van alle registers 4x256bit
signal reg_en: std_logic_vector(3 downto 0);
signal secnumber: std_logic_vector(2 downto 0);			--sectionnumber van het blokje (zie lane specs)
--signal bit_counter: std_logic_vector(63 downto 0); --waarom stond dit in comment? nodig?

--de toestanden van onze lane-fsm
type FSM_state is (S_IDLE, S_START, S_P0, S_P1, S_P2, S_P3, S_P4, S_P5, S_Q0, S_Q1, S_DONE);
signal state: FSM_state;

begin

--instantiaties van componenten

inst_expansion: Expansion
	port map(message_block, chaining_value, w(1535 downto 1280), w(1279 downto 1024), w(1023 downto 768),
		w(767 downto 512), w(511 downto 256), w(255 downto 0));
	
	
--de nieuwe implementatie heeft 4 256-bit registers nodig om de output van de permutatieblok bij te houden
gen_reg_256: for i in 3 downto 0 generate --zie figuur in thesis
	inst_reg256: reg256 
		port map(permut_out, clk, reg_en(i), reg_out((256*i + 255) downto 256*i));
end generate;


--ook hebben we een 256-bit mux met 5 ingangen nodig!
inst_mux256: mux256
	port map(w(1535 downto 1280), w(1279 downto 1024), w(1023 downto 768), w(767 downto 512), w(511 downto 256), w(255 downto 0), triexor(255 downto 0), mux_select(2 downto 0), mux_out(255 downto 0));


--ipv half en full rounds te genereren gaan we nu maar 1x (!!) een permutatieblok genereren, het rounds-signaal maakt verschil tussen P en Q
inst_round: round
	port map(mux_out, permut_out, rst, clk, start_round, done_round, bit_counter, secnumber, rounds);


--de combinatorische exors
triexor <= reg_out(1023 downto 768) xor reg_out(767 downto 512) xor reg_out(511 downto 256);
hash_out <= reg_out(511 downto 256) xor reg_out(255 downto 0);


done <= done_lane;

	
	fsm: process(rst, clk)
	begin
		if rst = '0' then
			reg_en <="0000";
			mux_select <= "000";	
			start_round <= '0';		
			rounds <= '0';			
			done_lane <= '0';		
			secnumber <= "000";
			state <= S_IDLE;
			
		elsif clk'event and clk = '1' then
			case state is
				when S_IDLE =>
					if start = '1' then
						reg_en <="0000";
						mux_select <= "000";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_START;
					else
						reg_en <="0000";
						mux_select <= "000";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_IDLE;
					end if;
					
				when S_START =>
						reg_en <="1000";
						mux_select <= "000";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_P0;
					
				when S_P0 =>
					if done_round = '1' then
						reg_en <="0100";
						mux_select <= "001";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "001";
						state <= S_P1;
					else
						reg_en <="1000";
						mux_select <= "000";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_P0;
					end if;
					
				when S_P1 =>
					if done_round = '1' then
						reg_en <="0010";
						mux_select <= "010";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "010";
						state <= S_P2;
					else
						reg_en <="0100";
						mux_select <= "001";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "001";
						state <= S_P1;
					end if;
				
				when S_P2 =>
					if done_round = '1' then
						reg_en <="0001";
						mux_select <= "110";
						start_round <= '1';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_Q0;
					else
						reg_en <="0010";
						mux_select <= "010";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "010";
						state <= S_P2;
					end if;
				
				when S_P3 =>
					if done_round = '1' then
						reg_en <="0100";
						mux_select <= "100";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "100";
						state <= S_P4;
					else
						reg_en <="1000";
						mux_select <= "011";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "011";
						state <= S_P3;
					end if;
					
				when S_P4 =>
					if done_round = '1' then
						reg_en <="0010";
						mux_select <= "101";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "101";
						state <= S_P5;
					else
						reg_en <="0100";
						mux_select <= "100";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "100";
						state <= S_P4;
					end if;
					
				when S_P5 =>
					if done_round = '1' then
						reg_en <="0000";
						mux_select <= "110";
						start_round <= '1';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "001";
						state <= S_Q1;
					else
						reg_en <="0010";
						mux_select <= "101";
						start_round <= '0';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "101";
						state <= S_P5;
					end if;
					
				when S_Q0 =>
					if done_round = '1' then
						reg_en <="1000";
						mux_select <= "011";
						start_round <= '1';
						rounds <= '1';
						done_lane <= '0';
						secnumber <= "011";
						state <= S_P3;
					else
						reg_en <="0001";
						mux_select <= "110";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_Q0;
					end if;
					
				when S_Q1 =>
					if done_round = '1' then
						reg_en <="0000";
						mux_select <= "000";
						start_round <= '1';
						rounds <= '0';
						done_lane <= '1';
						secnumber <= "000";
						state <= S_DONE;
					else
						reg_en <="0010";
						mux_select <= "110";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "001";
						state <= S_Q1;
					end if;
					
				when S_DONE =>
					if start = '1' then
						reg_en <="0000";
						mux_select <= "000";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '1';
						secnumber <= "000";
						state <= S_DONE;
					else
						reg_en <="0000";
						mux_select <= "000";
						start_round <= '0';
						rounds <= '0';
						done_lane <= '0';
						secnumber <= "000";
						state <= S_IDLE;
					end if;
					
				when others =>
					reg_en <="0000";
					mux_select <= "000";						
					start_round <= '0';
					rounds <= '0';
					done_lane <= '1';
					secnumber <= "000";
					state <= S_IDLE;
					
			end case;
		end if;
	end process;
	
end Behavioral;

